#include "journal.hpp"

#include <chrono>
#include <compare>
#include <concepts>
#include <filesystem>
#include <fstream>
#include <iostream>
#include <memory>
#include <mutex>
#include <optional>
#include <ostream>
#include <stop_token>
#include <string>
#include <string_view>
#include <thread>

namespace journal {
namespace mutex {
class Lock {
  std::mutex &mutex;
  bool moved = false;

public:
  Lock(Lock &rhs) = delete;
  Lock(std::mutex &mutex) : mutex(mutex) { mutex.lock(); 
  };
  Lock(Lock &&rhs) : mutex(rhs.mutex) { rhs.moved = true; 
  };
  auto operator=(Lock &rhs) = delete;
  ~Lock() {
    if (!moved) {
      mutex.unlock();
    }
  }
};

template <typename T> class Mutex {
  std::mutex mutex;
  T inner;

public:
  template <typename U>
  Mutex(U &&value) requires(std::assignable_from<U, T>)
      : inner(std::move(value)) {}
  Mutex() requires(std::default_initializable<T>) : inner(T{}) {}

  auto ref() -> std::pair<Lock, const T &> {
    return {std::move(scoped_lock()), inner};
  }
  auto mut() -> std::pair<Lock, T &> {
    return {std::move(scoped_lock()), inner};
  }

  auto operator->() -> T& {
    return inner;
  }

  auto scoped_lock() -> Lock { return Lock(mutex); }

  auto unique_lock() -> std::unique_lock<std::mutex> {
    return std::unique_lock(mutex);
  }

  auto lock() -> void { mutex.lock(); }

  auto try_lock() -> bool { return mutex.try_lock(); }

  auto unlock() -> void { mutex.unlock(); }
};
} // namespace mutex

namespace detail {

enum Color {
  Black,
  Red,
  Green,
  Yellow,
  Blue,
  Magenta,
  Cyan,
  White,
  BrightBlack,
  BrightRed,
  BrightGreen,
  BrightYellow,
  BrightBlue,
  BrightMagenta,
  BrightCyan,
  BrightWhite,
  None,
};

auto color_code(const Color &color) -> std::string {
  switch (color) {
  case Black:
    return "\e[0;30m";
  case Red:
    return "\e[0;31m";
  case Green:
    return "\e[0;32m";
  case Yellow:
    return "\e[0;33m";
  case Blue:
    return "\e[0;34m";
  case Magenta:
    return "\e[0;35m";
  case Cyan:
    return "\e[0;36m";
  case White:
    return "\e[0;37m";

  case BrightBlack:
    return "\e[1;30m";
  case BrightRed:
    return "\e[1;31m";
  case BrightGreen:
    return "\e[1;32m";
  case BrightYellow:
    return "\e[1;33m";
  case BrightBlue:
    return "\e[1;34m";
  case BrightMagenta:
    return "\e[1;35m";
  case BrightCyan:
    return "\e[1;36m";
  case BrightWhite:
    return "\e[1;37m";
  case None:
    return "\e[0;0m";
    break;
  }
}

auto colored(const std::string &str, const Color &color) -> std::string {
  return color_code(color) + str + color_code(Color::None);
}

constexpr auto to_string_colored(const LogVerbosity &verbosity) -> std::string {
  switch (verbosity) {
  case LogVerbosity::Trace:
    return colored("TRACE", White);
  case LogVerbosity::Debug:
    return colored("DEBUG", Magenta);
  case LogVerbosity::Info:
    return colored("INFO ", Green);
  case LogVerbosity::Warn:
    return colored("WARN ", Yellow);
  case LogVerbosity::Error:
    return colored("ERROR", Red);
  }
}

constexpr auto to_string(const LogVerbosity& verbosity) -> std::string {
    switch (verbosity) {
        case LogVerbosity::Trace:
          return "TRACE";
        case LogVerbosity::Debug:
          return "DEBUG";
        case LogVerbosity::Info:
          return "INFO ";
        case LogVerbosity::Warn:
          return "WARN ";
        case LogVerbosity::Error:
          return "ERROR";
        }
}

namespace formatter {

struct DefaultFormatter : public Formatter {
  virtual auto operator()(const Verbosity &verbosity,
                          const source_location &source) -> std::string {
    return std::format("[{}] [{: <5}] [{}]", get_current_time_string(),
                       to_string(verbosity.inner),
                       get_source_location_string(source));
  }

  auto get_source_location_string(const source_location &source)
      -> std::string {
    const auto file =
        std::filesystem::path(source.file_name()).filename().string();
    return std::format("{}::{}", file, source.line());
  }

  auto get_current_time_string() -> std::string {
    const auto now = std::chrono::utc_clock::now();

    return std::format("{:%Y-%m-%d %H:%M:%OS} UTC", now);
  }
};
} // namespace formatter

using namespace std::chrono_literals;

namespace sinks {

struct BaseSink {
  Verbosity min_verbosity;
  Verbosity max_verbosity;

  BaseSink()
      : min_verbosity(LogVerbosity::Trace), max_verbosity(LogVerbosity::Error) {
  }

  BaseSink(const Verbosity &min, const Verbosity &max)
      : min_verbosity(min), max_verbosity(max) {}

  auto should_sink_verbosity(const Verbosity &verbosity) const -> bool {
    return verbosity >= min_verbosity && verbosity <= max_verbosity;
  }

  auto set_max_verbosity(const Verbosity &new_max) -> Verbosity {
    return std::exchange(max_verbosity, new_max);
  }

  auto set_min_verbosity(const Verbosity &new_min) -> Verbosity {
    return std::exchange(min_verbosity, new_min);
  }

  virtual auto flush() -> void {}
  virtual auto operator<<(std::string_view str) -> void {}
};

struct StdOutSink : public BaseSink {
  StdOutSink() : BaseSink(LogVerbosity::Trace, LogVerbosity::Info) {}

  auto write(std::string_view str) -> void { (*this) << str; }
  virtual auto flush() -> void { std::cout.flush(); }
  virtual auto operator<<(std::string_view str) -> void { std::cout << str; }
};

struct StdErrSink : public BaseSink {
  StdErrSink() : BaseSink(LogVerbosity::Warn, LogVerbosity::Error) {}

  virtual auto flush() -> void { std::cerr.flush(); }
  virtual auto operator<<(std::string_view str) -> void { std::cerr << str; }
};

struct OStreamSink : public BaseSink {
  std::shared_ptr<std::ostream> ostream;

  OStreamSink(const std::shared_ptr<std::ostream> &stream,
              const Verbosity &min = LogVerbosity::Trace,
              const Verbosity &max = LogVerbosity::Error)
      : ostream(stream), BaseSink(min, max) {}

  virtual auto flush() -> void { ostream->flush(); }
  virtual auto operator<<(std::string_view str) -> void { (*ostream) << str; }
};

struct FileSink : public OStreamSink {
  FileSink(const std::string &file_path,
           std::ios::ios_base::openmode open_mode = std::ios::ios_base::app,
           const Verbosity &min = LogVerbosity::Trace,
           const Verbosity &max = LogVerbosity::Error)
      : OStreamSink(std::make_shared<std::ofstream>(file_path, open_mode), min,
                    max) {}

  FileSink(const std::shared_ptr<std::ofstream> &file,
           const Verbosity &min = LogVerbosity::Trace,
           const Verbosity &max = LogVerbosity::Error)
      : OStreamSink(file, min, max) {}
};

} // namespace sinks

static auto try_flush_all_loop(std::stop_token) -> void;

class Journal {
  std::vector<std::unique_ptr<sinks::BaseSink>> sinks;
  std::unique_ptr<formatter::Formatter> formatter;
  std::optional<std::jthread> flush_thread;
  std::chrono::system_clock::duration flush_interval;

public:
  Journal()
      : flush_interval(250ms),
        formatter(std::make_unique<formatter::DefaultFormatter>()) {}
  ~Journal() {
    if (flush_thread) {
      log(LogVerbosity::Trace, source_location::current(),
          "Requesting stop for flush thread");
      flush_thread->request_stop();
      log(LogVerbosity::Trace, source_location::current(),
          "Waiting for flush thread to join..");
      flush_thread->join();
      log(LogVerbosity::Trace, source_location::current(),
          "Flush thread joined");
    }
    flush_all();
  }

  auto reset_formatter() -> void {
    formatter = std::make_unique<formatter::DefaultFormatter>();
  }

  auto set_formatter(std::unique_ptr<formatter::Formatter> &&formatter)
      -> void {
    formatter = std::move(formatter);
  }

  auto
  set_flush_interval(const std::chrono::system_clock::duration &new_interval)
      -> void {
    flush_interval = new_interval;
  }
  auto get_flush_interval() const -> std::chrono::system_clock::duration {
    return flush_interval;
  }

  auto init_flush_thread() -> void {
    flush_thread = std::jthread(&detail::try_flush_all_loop);
  }

  auto should_flush() const -> bool {
    return !flush_thread.has_value() || flush_interval < 1ms;
  }

  auto flush_all() -> void { force_flush_all(); }

  auto force_flush_all() -> void {
    for (auto &&sink : sinks) {
      sink->flush();
    }
  }

  auto log(const Verbosity &verbosity, source_location source,
           std::string_view msg) -> void {
    const auto do_flush = should_flush();
    for (auto &&sink : sinks) {
      if (sink->should_sink_verbosity(verbosity)) {
        (*sink) << format(msg, verbosity, source);

        if (do_flush) {
          sink->flush();
        }
      }
    }
  }

auto format(std::string_view str, const Verbosity &verbosity,
            const source_location &source) -> std::string {
  return std::format("{} {}\n", (*formatter)(verbosity, source), str);
}

auto add_file(const std::string &file_name,
              std::ios::ios_base::openmode open_mode = std::ios::ios_base::app,
              const Verbosity &min = LogVerbosity::Trace,
              const Verbosity &max = LogVerbosity::Error) -> void {
  auto file = std::make_shared<std::ofstream>(file_name, open_mode);
  if (file->is_open() && !(file->bad() || file->fail())) {
    add_sink(std::make_unique<sinks::FileSink>(file, min, max));
  }
}

auto add_sink(std::unique_ptr<sinks::BaseSink> &&sink) -> void {
  sinks.emplace_back(std::move(sink));
}

auto add_std_out() { add_sink(std::make_unique<sinks::StdOutSink>()); }
auto add_std_err() { add_sink(std::make_unique<sinks::StdErrSink>()); }
};

static mutex::Mutex<std::unique_ptr<Journal>> global_journal;

static auto try_flush_all_loop(std::stop_token stoken) -> void {
  std::chrono::system_clock::duration interval = 250ms;
  while (!stoken.stop_requested()) {
    {
      auto &&[lock, ref] = global_journal.mut();
      interval = ref->get_flush_interval();
      ref->flush_all();
    }
    std::this_thread::sleep_for(interval);
  }
}

auto log_internal(const LogVerbosity &verbosity, detail::source_location source,
                  const std::string &msg) -> void {
  auto &&[lock, ref] = global_journal.ref();

  ref->log(verbosity, source, msg);
}
} // namespace detail

auto init() -> void {
  auto &&[lock, ref] = detail::global_journal.mut();
  if (!ref.get()) {
    ref = std::make_unique<detail::Journal>();
    ref->init_flush_thread();
  }
}

auto init_no_threads() -> void {
  auto &&[lock, ref] = detail::global_journal.mut();
  if (!ref.get()) {
    ref = std::make_unique<detail::Journal>();
  }
}

auto init_std_out(const LogVerbosity &min) -> void {
  init();
  auto &&[lock, ref] = detail::global_journal.mut();
  ref->add_std_out();
}

auto init_std_out_err(const LogVerbosity &min) -> void {
  init();
  auto &&[lock, ref] = detail::global_journal.mut();
  ref->add_std_out();
  ref->add_std_err();
}

auto set_formatter(std::unique_ptr<detail::formatter::Formatter> &&formatter)
    -> void {
  auto &&[lock, ref] = detail::global_journal.mut();
  ref->set_formatter(std::move(formatter));
}

auto add_std_out() -> void {
  auto &&[lock, ref] = detail::global_journal.mut();
  ref->add_std_out();
}
auto add_std_err() -> void {
  auto &&[lock, ref] = detail::global_journal.mut();
  ref->add_std_err();
}

auto add_sink(std::shared_ptr<std::ostream> stream, const LogVerbosity &min,
              const LogVerbosity &max) -> void {
  auto &&[lock, ref] = detail::global_journal.mut();
  ref->add_sink(std::make_unique<detail::sinks::OStreamSink>(stream, min, max));
}

auto add_file(std::string_view file_path, std::ios_base::openmode open_mode,
              const LogVerbosity &min, const LogVerbosity &max) -> void {
  auto &&[lock, ref] = detail::global_journal.mut();
  ref->add_file(std::string(file_path), open_mode, min, max);
}
} // namespace journal
